import traceback
from pathlib import Path
import shutil
from uuid import uuid4
import os

class TempFile:
    def __init__(self, path: Path, dont_delete: bool):
        self.path = path
        self.dont_delete = dont_delete

    def __enter__(self):
        os.makedirs(self.path.parent, exist_ok=True)
        if self.path.is_file():
            os.unlink(self.path)
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.dont_delete:
            if self.path.is_file():
                try:
                    os.unlink(self.path)
                except:
                    print("Cannot delete test file:\n"+traceback.format_exc())


class TempFolder:
    def __init__(self, path: Path, dont_delete: bool = False):
        self.path = path
        self.dont_delete = dont_delete

    def __enter__(self):
        if self.path.is_dir():
            shutil.rmtree(self.path)
        os.makedirs(self.path)
        return self.path

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.dont_delete and self.path.is_dir():
            shutil.rmtree(self.path, ignore_errors=True)


class Locator:
    def __init__(self, root_path: Path|None = None):
        if root_path is None:
            root_path = Path(__file__).parent.parent.parent
        self._root_path = root_path

    def _make_and_return(self, path) -> Path:
        os.makedirs(path, exist_ok=True)
        return path

    @property
    def data_folder(self) -> Path:
        return self._make_and_return(self._root_path/'data')

    @property
    def resources_folder(self) -> Path:
        return self._make_and_return(self._root_path/'data/resources')

    @property
    def cache_folder(self) -> Path:
        return self._make_and_return(self._root_path/'temp/brainbox_cache')

    @property
    def temp_folder(self) -> Path:
        return self._make_and_return(self._root_path/'temp')

    @property
    def test_folder(self) -> Path:
        return self._make_and_return(self._root_path/'temp/tests')

    @property
    def self_test_path(self) -> Path:
        return self._make_and_return(self._root_path/'data/brainbox_self_test')

    @property
    def db_path(self) -> Path:
        return self.data_folder/'brainbox.db'

    @property
    def root_folder(self) -> Path:
        return self._root_path


    def create_test_file(self, extension_without_leading_dot: str|None = None, subfolder: str|None = None, dont_delete: bool = False) -> TempFile:
        path = self.test_folder
        if subfolder is not None:
            path /= subfolder
        name = str(uuid4())
        if extension_without_leading_dot is not None:
            name+= '.'+extension_without_leading_dot
        path/=name
        return TempFile(path, dont_delete)

    def create_test_folder(self, subfolder: str|None = None, dont_delete: bool = False) -> TempFolder:
        path = self.test_folder
        if subfolder is not None:
            path /= subfolder
        path /= str(uuid4())
        return TempFolder(path, dont_delete)

Loc = Locator()



import datetime
import re
import time
import requests

class ApiUtils:
    @staticmethod
    def check_address(address):
        if not re.match('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}:\d{1,5}', address):
            raise ValueError(f'Address must be IP:port, no protocol, no trailing slash, but was:\n{address}')


    @staticmethod
    def wait_for_reply(url, time_in_seconds, endpoint_name=''):
        reply = None
        if endpoint_name == '':
            endpoint_name = url
        begin = datetime.datetime.now()
        for i in range(time_in_seconds * 100):
            time.sleep(0.01)
            if i>2 and (datetime.datetime.now()-begin).total_seconds() > time_in_seconds:
                break
            try:
                reply = requests.get(url)
            except:
                continue
            if reply.status_code == 200:
                break
        if reply is None:
            raise ValueError(
                f"Endpoint {endpoint_name} was not reacheable within {time_in_seconds} seconds")
        if reply.status_code != 200:
            raise ValueError(
                f"Endpoint {endpoint_name} was reacheable, but returned bad status code {reply.status_code}\n{reply.text}")





from .endpoint import endpoint
from .server import Server
from .api import Api, bind_to_api
from .test_api import TestApi
from .api_utils import ApiUtils
from .signature_processor import SignatureProcessor


import pickle

def try_pickle(obj, path='root'):
    try:
        pickle.dumps(obj)
        return None
    except Exception as e:
        return f"{path}: {e}"

def find_pickle_errors(obj, path='root', seen=None):
    if seen is None:
        seen = set()
    errors = []

    obj_id = id(obj)
    if obj_id in seen:
        return errors
    seen.add(obj_id)

    err = try_pickle(obj, path)
    if err is None:
        return errors  # объект сериализуется

    # Раскрываем составные объекты
    if isinstance(obj, dict):
        for k, v in obj.items():
            errors += find_pickle_errors(k, f"{path}.key({repr(k)})", seen)
            errors += find_pickle_errors(v, f"{path}[{repr(k)}]", seen)
    elif isinstance(obj, (list, tuple, set)):
        for i, item in enumerate(obj):
            errors += find_pickle_errors(item, f"{path}[{i}]", seen)
    elif hasattr(obj, '__dict__'):
        for attr_name, attr_value in vars(obj).items():
            errors += find_pickle_errors(attr_value, f"{path}.{attr_name}", seen)
    elif hasattr(obj, '__slots__'):
        for slot in obj.__slots__:
            try:
                attr_value = getattr(obj, slot)
                errors += find_pickle_errors(attr_value, f"{path}.{slot}", seen)
            except AttributeError:
                continue
    else:
        # Просто не сериализуемый примитив
        errors.append(f"{path}: {type(obj).__name__} is not picklable")

    return errors


from jinja2 import Template, Environment, meta
from typing import Generic
from .abstract_prompter import IPrompter, T
from dataclasses import is_dataclass
from copy import copy
import re
class JinjaPrompter(IPrompter, Generic[T]):
    def __init__(self, template: str, prettify_newlines: bool = True):
        self._template_text = template.replace('<<', '{{').replace('>>', '}}')
        self._template = Template(self._template_text)
        self._prettify_newlines = prettify_newlines

    @staticmethod
    def normalize_newlines(text: str) -> str:
        text = re.sub(r'\n{2,}', '\n\n', text)
        text = re.sub(r'(?<!\n)\n(?!\n)', ' ', text)
        return text

    def get_variables(self):
        env = Environment()
        parsed_content = env.parse(self._template_text)
        variables = meta.find_undeclared_variables(parsed_content)
        return variables

    def __call__(self, obj: T) -> str:
        values = {}
        if is_dataclass(obj):
            if isinstance(obj, type):
                pass
            else:
                values = copy(obj.__dict__)
        elif isinstance(obj, dict):
            values = copy(obj)
        values['_'] = obj
        s = self._template.render(**values)

        if self._prettify_newlines:
            s = JinjaPrompter.normalize_newlines(s)

        return s





from typing import TypeVar, Generic, Callable
from .server import Server
from .api import Api
from ..fork import Fork


TApi = TypeVar('TApi')

class TestApi(Generic[TApi]):
    def __init__(self,
                 api_factory: Callable[[str], Api],
                 server: Server
                 ):
        self.api_factory = api_factory
        self.server = server
        self.fork = None

    def __enter__(self) -> TApi:
        self.fork = Fork(self.server).start()
        api = self.api_factory(f'127.0.0.1:{self.server.port}')
        api.wait()
        return api

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.fork is not None:
            self.fork.terminate()





from .abstract_prompter import IPrompter, T, Generic
from .address import Address
from .template_parts import ITemplatePart, ConstantTemplatePart
import re

class Parser:
    def __init__(self):
        self.parts = []

    def parse_next(self, part: str):
        if part.startswith('<<') and part.endswith('>>'):
            uid = part[2:-2]
            token = ITemplatePart.parse_gc(uid)
            self.parts.append(token)
        else:
            if part != '':
                self.parts.append(ConstantTemplatePart(part))

    def finalize(self):
        return tuple(self.parts)



def _parse(template: str) -> tuple[ITemplatePart,...]:
    parser = Parser()
    parts = re.split('(<<[^>]+>>)', template)
    for p in parts:
        parser.parse_next(p)
    return parser.finalize()


class Prompter(IPrompter[T], Generic[T]):
    def __init__(self, template: str|tuple[ITemplatePart,...]):
        if isinstance(template, str):
            self.template = _parse(template)
        else:
            self.template = template

    def __call__(self, obj: T) -> str:
        return ''.join(part.to_str(obj) for part in self.template)

    def to_readable_string(self):
        return ''.join(p.to_readable_expression() for p in self.template)




from abc import ABC, abstractmethod

class IAddressElement(ABC):
    @abstractmethod
    def get(self, obj):
        pass

    @abstractmethod
    def set(self, obj, value):
        pass

    @staticmethod
    def translate(element) -> 'IAddressElement':
        if isinstance(element, IAddressElement):
            return element
        return DefaultElement(element)



class DefaultElement(IAddressElement):
    def __init__(self, element):
        self.element = element

    def get(self, obj):
        if isinstance(self.element, str) and hasattr(obj, self.element):
            return getattr(obj, self.element)
        if hasattr(obj, '__getitem__'):
            getitem = getattr(obj, '__getitem__')
            if callable(getitem):
                return getitem(self.element)
        raise ValueError(f"Obj {obj} cannot be addressed with {self.element}")

    def set(self, obj, value):
        if hasattr(obj, '__setitem__'):
            setitem = getattr(obj, '__setitem__')
            if callable(setitem):
                setitem(obj, self.element, value)
                return
        if isinstance(self.element, str):
            setattr(obj, self.element, value)
            return
        raise ValueError(f"Obj {obj} cannot be set with address `{self.element}`")

    def __str__(self):
        if isinstance(self.element, str):
            return self.element
        return f'[{self.element}]'


class Address:
    def __init__(self, *address):
        from .address_builder import AddressBuilder
        if len(address) == 1:
            if isinstance(address[0], AddressBuilder):
                self.address = address[0]._address_builder_stored_address.address
            elif isinstance(address[0], Address):
                self.address = address[0].address
            else:
                self.address = (IAddressElement.translate(address[0]), )
        else:
            self.address: tuple[IAddressElement,...] = tuple(IAddressElement.translate(a) for a in address)

    def get(self, obj):
        result = obj
        for index, element in enumerate(self.address):
            try:
                result = element.get(result)
            except Exception as exc:
                raise ValueError(f"Address {self} failed at {index} for object\n{obj}") from exc
        return result

    def set(self, obj, value):
        obj = self.pop().get(obj)
        self.address[-1].set(obj, value)

    def append(self, value: str|IAddressElement):
        if isinstance(value, str):
            value= IAddressElement.translate(value)
        elif isinstance(value, IAddressElement):
            pass
        else:
            raise ValueError("Expected string or IAddressElement")
        return Address(*(self.address+(value,)))

    def pop(self) -> 'Address':
        if len(self.address) == 0:
            raise ValueError("Can't pop from empty address")
        return Address(*self.address[:-1])

    def is_empty(self):
        return len(self.address) == 0

    def __str__(self):
        return '.'.join(str(a) for a in self.address)

    @staticmethod
    def parse(s: str):
        parts = s.split('.')
        parts = [DefaultElement(e) for e in parts]
        return Address(*parts)



from dataclasses import dataclass
from typing import *
from abc import ABC, abstractmethod
from .address_builder import AddressBuilderGC
from .address import Address
from .abstract_prompter import IPrompter
from .referrer import Referrer

class ITemplatePart(ABC):
    @abstractmethod
    def to_str(self, value: Any) -> str:
        pass

    @abstractmethod
    def to_readable_expression(self) -> str:
        pass

    @staticmethod
    def parse_gc(uid: str):
        address = AddressBuilderGC.find(AddressBuilderGC.Dimension.address, uid)
        if address is None:
            raise ValueError(f"Cannot find uid {uid} in the cache")
        if AddressBuilderGC.find(AddressBuilderGC.Dimension.subprompt, uid):
            return SubpromptPropagationTemplatePart(address)
        else:
            return AddressTemplatePart(
                address,
                AddressBuilderGC.find(AddressBuilderGC.Dimension.operator, uid),
                AddressBuilderGC.find(AddressBuilderGC.Dimension.misc, uid),
            )


@dataclass
class ConstantTemplatePart(ITemplatePart):
    value: str

    def to_str(self, value: Any) -> str:
        return self.value

    def to_readable_expression(self) -> str:
        return self.value

@dataclass
class AddressTemplatePart(ITemplatePart):
    address: Address
    formatter: Callable[[Any], str]|None = None
    misc: Any = None

    def to_str(self, value: Any) -> str:
        result = self.address.get(value)
        if self.formatter is not None:
            result = self.formatter(result)
        return str(result)

    def to_readable_expression(self) -> str:
        return '{{'+self.address.__str__()+'}}'

@dataclass
class SubpromptPropagationTemplatePart(ITemplatePart):
    address: Address

    def to_str(self, value: Any):
        return self.address.get(value)(value)

    def to_readable_expression(self) -> str:
        return '{{'+self.address.__str__()+"(_)}}"




import inspect
from dataclasses import dataclass
from collections import OrderedDict

@dataclass
class SignatureProcessor:
    mandatory: tuple[str,...]
    optional: tuple[str,...]
    open: bool

    def to_kwargs_only(self, *args, **kwargs):
        all = self.mandatory + self.optional
        if len(args) > len(all):
            raise ValueError(f"args has {len(args)} arguments, while there are only {len(all)} arguments")
        result = OrderedDict()
        for i, arg in enumerate(args):
            name = all[i]
            if name in kwargs:
                raise ValueError(f"Args at index {i} conflicts with kwargs {name}")
            result[name] = arg

        for key, value in kwargs.items():
            result[key] = value

        mandatory_seen = set()
        for key in result:
            if key in self.mandatory:
                mandatory_seen.add(key)
            elif key in self.optional:
                pass
            else:
                if not open:
                    raise ValueError(f"Argument `{key}` is not in fields, and the signature is not open")

        if len(mandatory_seen) < len(self.mandatory):
            not_seen = ", ".join([f'`{c}`' for c in self.mandatory if c not in mandatory_seen])
            raise TypeError(f"Not all mandatory arguments are provided: missing {not_seen}")

        return result

    @staticmethod
    def from_signature(method):
        signature = inspect.signature(method)
        mandatory = list()
        optional = list()
        open = False
        for i,(p,v) in enumerate(signature.parameters.items()):
            if i == 0 and v.name=='self':
                continue
            if v.kind == inspect.Parameter.VAR_KEYWORD:
                open = True
            else:
                if v.default == inspect._empty:
                    mandatory.append(v.name)
                else:
                    optional.append(v.name)
        return SignatureProcessor(tuple(mandatory), tuple(optional), open)


from typing import *
from .address import DefaultElement, Address
from uuid import uuid4
from enum import IntEnum


class AddressBuilderGC:
    class Dimension(IntEnum):
        address = 0
        operator = 1
        subprompt = 2
        misc = 100

    cache: dict['AddressBuilderGC.Dimension', dict[str,Any]] = {}

    @staticmethod
    def record(dimension: 'AddressBuilderGC.Dimension', id: str, value: Any):
        if dimension not in AddressBuilderGC.cache:
            AddressBuilderGC.cache[dimension] = {}
        AddressBuilderGC.cache[dimension][id] = value

    @staticmethod
    def find(dimension: 'AddressBuilderGC.Dimension', id: str):
        if dimension not in AddressBuilderGC.cache:
            return None
        if id not in AddressBuilderGC.cache[dimension]:
            return None
        return AddressBuilderGC.cache[dimension][id]


RESERVED_FIELDS = {
    '_address_builder_stored_address',
    '_address_builder_uuid',
    '__str__',
    '__truediv__',
}

class AddressBuilder:
    def __init__(self, address: Address|None = None):
        self._address_builder_uuid = 'id'+str(uuid4()).replace('-','')
        self._address_builder_stored_address = address if address is not None else Address()
        AddressBuilderGC.record(
            AddressBuilderGC.Dimension.address,
            self._address_builder_uuid,
            self._address_builder_stored_address
        )

    def __str__(self):
        return f'<<{self._address_builder_uuid}>>'

    def __getattribute__(self, item):
        if item in RESERVED_FIELDS:
            return super().__getattribute__(item)
        return AddressBuilder(self._address_builder_stored_address.append(DefaultElement(item)))

    def __truediv__(self, other):
        AddressBuilderGC.record(
            AddressBuilderGC.Dimension.operator,
            self._address_builder_uuid,
            other
        )
        return self.__str__()

    def __invert__(self):
        AddressBuilderGC.record(
            AddressBuilderGC.Dimension.subprompt,
            self._address_builder_uuid,
            True
        )
        return self.__str__()





from .abstract_prompter import IPrompter
from .address import Address, IAddressElement, DefaultElement
from .address_builder import AddressBuilder, AddressBuilderGC
from .prompter import Prompter
from .template_parts import ITemplatePart, ConstantTemplatePart, AddressTemplatePart, SubpromptPropagationTemplatePart
from .referrer import Referrer
from .jinja_prompter import JinjaPrompter



from typing import *
from dataclasses import dataclass
from .signature_processor import SignatureProcessor
from .endpoint import EndpointMetadata

@dataclass
class MarshallingMetadata:
    name: str
    method: Callable
    endpoint: EndpointMetadata
    signature: SignatureProcessor

    def get_endpoint_address(self):
        if self.endpoint.url is None:
            return '/'+self.name
        return self.endpoint.url

    @staticmethod
    def _get_endpoints(_obj, _type) -> tuple['MarshallingMetadata',...]:
        result = []
        for attr_name, attr_value in _type.__dict__.items():
            if not hasattr(attr_value,'_metadata'):
                continue
            result.append(MarshallingMetadata(
                attr_name,
                getattr(_obj, attr_name) if _obj is not None else None,
                attr_value._metadata,
                SignatureProcessor.from_signature(attr_value)
            ))
        return tuple(result)

    @staticmethod
    def get_endpoints_from_object(obj) -> tuple['MarshallingMetadata',...]:
        return MarshallingMetadata._get_endpoints(obj, type(obj))

    @staticmethod
    def get_endpoints_from_type(type: Type) -> tuple['MarshallingMetadata',...]:
        return MarshallingMetadata._get_endpoints(None, type)


import sys
import pickle
import os
import traceback

if __name__ == '__main__':
    name = sys.argv[1]
    file = sys.argv[2]

    try:
        with open(file, 'rb') as stream:
            method = pickle.load(stream)
        os.unlink(file)
    except:
        print(f"Subprocess for {name}: cannot read entry point", file=sys.stderr)
        print(traceback.format_exc(), file=sys.stderr)
        raise

    try:
        method()
    except:
        print(f"Subprocess for {name}: exception in entry point", file=sys.stderr)
        print(traceback.format_exc(), file=sys.stderr)
        raise





from typing import *
from dataclasses import dataclass
from enum import Enum
from datetime import time, datetime, timedelta, date
from abc import ABC, abstractmethod


class Weekdays(Enum):
    Monday = 0
    Tuesday = 1
    Wednesday = 2
    Thursday = 3
    Friday = 4
    Saturday = 5
    Sunday = 6


class IRepetition(ABC):
    @abstractmethod
    def accepts_date(self, date: date) -> bool:
        pass

    def get_acceptable_dates(self, begin: datetime, end: datetime) -> Iterator[date]:
        begin_date = datetime(begin.year, begin.month, begin.day)
        while True:
            if begin_date > end:
                break
            if begin_date >= begin:
                if self.accepts_date(begin_date.date()):
                    yield begin_date.date()
            begin_date = begin_date + timedelta(days=1)



class WeekdayRepetition(IRepetition):
    def __init__(self, days: Weekdays | int | Iterable[Weekdays|int]):
        try:
            iterator = iter(days)
            iterable = True
        except:
            iterable = False
        if not iterable:
            days = [days]
        parsed = []
        for i, day in enumerate(days):
            if isinstance(day, int):
                parsed.append(Weekdays(day))
            elif isinstance(day, Weekdays):
                parsed.append(day)
            else:
                raise ValueError(f"Argument {i} should be Weekday or int, but was {day}")
        self._days = tuple(c.value for c in parsed)


    def accepts_date(self, date: date) -> bool:
        return date.weekday() in self._days


class EveryDayRepetition(IRepetition):
    def accepts_date(self, date: date) -> bool:
        return True


class CombinedRepetition(IRepetition):
    def __init__(self, repetitions: Iterable[IRepetition]):
        self.repetitions = tuple(repetitions)

    def accepts_date(self, date: date) -> bool:
        for rep in self.repetitions:
            if not rep.accepts_date(date):
                return False
        return True



@dataclass
class ScheduledTime:
    time: time
    duration: timedelta | None = None
    repetition: IRepetition = EveryDayRepetition()

    def all_intersections_with_interval(self, begin: datetime, end: datetime) -> Iterator[datetime]:
        if begin > end:
            raise ValueError(f'`begin` must not be greater than `end`, but were: {begin}, {end}')
        begin_1 = datetime(begin.year, begin.month, begin.day, self.time.hour, self.time.minute, self.time.second)
        while True:
            if not self.repetition.accepts_date(begin_1.date()):
                begin_1 += timedelta(days=1)
                continue
            if begin_1 > end:
                break
            if self.duration is None:
                if begin_1 >= begin and begin_1 <= end:
                    yield begin_1
            else:
                end_1 = begin_1 + self.duration
                if begin_1 <= end and end_1 >= begin:
                    yield begin_1
            begin_1 += timedelta(days=1)

    def intersects_with_interval(self, begin: datetime, end: datetime) -> bool:
        for _ in self.all_intersections_with_interval(begin, end):
            return True
        return False








    





import json
import pickle
import base64
import jsonpickle



class NoTupleJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, tuple):
            raise TypeError("Tuples are not allowed in JSON serialization.")
        return super().default(obj)


class Format:
    CONTROL_FIELD = '@type'
    CONTENT_FIELD = '@content'

    @staticmethod
    def check_json(obj):
        if obj is None:
            return True
        if isinstance(obj, str) or isinstance(obj, int) or isinstance(obj, float) or isinstance(obj, bool):
            return True
        if isinstance(obj, list):
            return all(Format.check_json(z) for z in obj)
        if isinstance(obj, dict):
            return all(Format.check_json(z) for z in obj.values())
        return False

    @staticmethod
    def encode(data: dict, use_json_pickle: bool = False):
        result = {}
        for key, value in data.items():
            if Format.check_json(value):
                result[key] = value
            else:
                if use_json_pickle:
                    result[key] = {
                        Format.CONTROL_FIELD: 'jsonpickle',
                        Format.CONTENT_FIELD: json.loads(jsonpickle.dumps(value))
                    }
                else:
                    s = base64.b64encode(pickle.dumps(value)).decode('ascii')
                    result[key] = {
                        Format.CONTROL_FIELD: 'base64', Format.CONTENT_FIELD: s
                    }
        return result

    @staticmethod
    def decode(data: dict):
        result = {}
        for key in data:
            if (
                isinstance(data[key], dict) and
                Format.CONTROL_FIELD in data[key] and
                Format.CONTENT_FIELD in data[key]
            ):
                if data[key][Format.CONTROL_FIELD] == 'base64':
                    result[key] =  pickle.loads(base64.b64decode(data[key][Format.CONTENT_FIELD]))
                elif data[key][Format.CONTROL_FIELD] == 'jsonpickle':
                    result[key] = jsonpickle.loads(json.dumps(data[key][Format.CONTENT_FIELD]))
                else:
                    raise ValueError("only @type=base64 and @type=jsonpickle is supported")
            else:
                result[key] = data[key]
        return result



from .address_builder import AddressBuilder
from typing import TypeVar, Generic

T = TypeVar('T')

class Referrer(Generic[T]):
    @property
    def ref(self) -> T:
        return AddressBuilder()

    def list_to_bullet_points(self, bullet_point = '* '):
        def _f(value):
            return '\n'.join(bullet_point+s for s in value)
        return _f






from .marshalling_metadata import MarshallingMetadata
from .format import Format
import flask
import traceback


class ServerBinding:
    def __init__(self,
                 meta :MarshallingMetadata,
                 custom_method_name: str|None = None
                 ):
        self.meta = meta
        self.custom_method_name = custom_method_name

    @property
    def __name__(self):
        if self.custom_method_name is None:
            return self.meta.name
        return self.custom_method_name

    def _get_arguments(self, kwargs, data):
        arguments = Format.decode(data['arguments'])
        for key, value in kwargs.items():
            if key in arguments:
                raise ValueError(f"{key} is provided via address and via json")
            arguments[key] = value
        return arguments

    def _process(self, kwargs, arguments):
        arguments = self._get_arguments(kwargs, arguments)
        result = self.meta.method(**arguments)
        result = dict(result=result, error = None)
        result = Format.encode(result, self.meta.endpoint.json_pickle_result)
        return result

    def __call__(self, **kwargs):
        data = flask.request.json
        try:
            return flask.jsonify(self._process(kwargs, data))
        except:
            tb = traceback.format_exc()
            print(tb)
            return flask.jsonify(dict(result=None, error=tb)), 500











from typing import *
from dataclasses import dataclass
from .fork import Fork

@dataclass
class ForkApp:
    main: Optional[Callable] = None
    supporting: Tuple[Callable,...] = ()
    _forks = []

    def run(self):
        for service in self.supporting:
            self._forks.append(Fork(service).start())
        if self.main is not None:
            self.main()

    def terminate(self):
        for fork in reversed(self._forks):
            fork.terminate()

    def __enter__(self):
        if self.main is not None:
            raise ValueError("Can only enter/exit in the app without main routine")
        self.run()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate()



